"""
oneline.core is the set of main functions and features, which contains:

    1. OneData, extended DataFrame
    2. OneSeries, extended Series
    3. plot(private), a plot function for fast using
"""

from .onedata import OneData
from .oneseries import OneSeries
