"""
The IO module supplies some basic IO methods with a more comprehensive auto detection. It's easy to use
without learning cost.

This module are still in progress.
"""

from .io import *
from .file_input import *
