"""
A set of methods that can make conversion related codes easier, faster and safer. This module is based on
built-in md5 and base64, and some of the methods can prevent memory overflow.

This module are still in progress.
"""

from .hash import *
